/* 
 * $RCSfile: idmap.c,v $
 *
 * x-kernel v3.2
 *
 * Copyright (c) 1993,1991,1990  Arizona Board of Regents
 *
 *
 * $Log: idmap.c,v $
 * Revision 1.38.2.1  1994/06/08  23:11:55  davidm
 * Reorganization ... key-specific routines are in idmap_special.h
 *
 * Revision 1.38  1993/12/11  00:23:33  menze
 * fixed #endif comments
 *
 * Revision 1.37  1993/12/07  02:25:12  menze
 * generichash was sometimes looking at too many bytes past the end of
 * the key.
 *
 */
#include "platform.h"
#include "idmap.h"

#include "upi.h"
#include "idmap_internal.h"
#include "xk_assert.h"
#include "xk_debug.h"

static struct {
    XMapResolveFunc	resolve;
    XMapBindFunc	bind;
    XMapUnbindFunc	unbind;
    XMapRemoveFunc	remove;
} map_functions[MAX_MAP_KEY_SIZE];

int	traceidmap;

#define MAX_MAP_SIZE 1024

static int		generichash (char *, int, int);
static xkern_return_t	mgenericresolve (Map, VOID *, VOID **);
static Bind 		mgenericbind (Map, VOID *, VOID *);
static xkern_return_t  	mgenericunbind (Map, VOID *);
static xkern_return_t  	mgenericremove (Map, Bind);
static xkern_return_t  	mgenericerror (Map, VOID *, VOID **);
static void		removeList (MapElement *);


/*
 * Returns ceil(log_2(x)).
 */
static int
ceil_log2(x)
     unsigned int x;
{
    int y = 0;

    while (x) {
	x >>= 1;
	++y;
    } /* while */
    return y;
} /* ceil_log2 */


/*
 * Create and return a new map containing a table with nEntries
 * entries mapping keys of size keySize to integers:
 */
Map
mapCreate(nEntries, keySize)
     int nEntries;
     int keySize;
{
    register Map m;
    register int log2;
    int entries = nEntries;

    xTrace1(idmap, TR_FULL_TRACE, "mapCreate called, %d entries", nEntries);
    xAssert(nEntries);
    log2 = ceil_log2(nEntries);
    xAssert(log2 < 8*sizeof(int) - 1);
    if (entries == 1<<(log2-1)) {
	nEntries = entries;		/* nEntries is a power of 2 */
    } else {
	nEntries = 1<<log2;		/* use next higher power of two */
    } /* if */
    xAssert(nEntries <= MAX_MAP_SIZE);

    xTrace1(idmap, TR_EVENTS, "mapCreate: actual map size %d\n", nEntries);
    if (keySize > MAX_MAP_KEY_SIZE) {
	return (Map) -1;
    }
    m = (Map) xMalloc(sizeof(*m));
    m->tableSize = nEntries;
    m->keySize = keySize;
    m->cache = 0;
    m->freelist = 0;
    m->table = (MapElement **)xMalloc(nEntries * sizeof(MapElement *));
    bzero((char *)m->table, nEntries * sizeof(MapElement *));
    if (keySize > 0) {
	if (map_functions[keySize].resolve != 0) {
	    m->resolve = map_functions[keySize].resolve;
	    m->bind    = map_functions[keySize].bind   ;
	    m->unbind  = map_functions[keySize].unbind ;
	    m->remove  = map_functions[keySize].remove;
	} else {
	    m->resolve = mgenericresolve;
	    m->bind    = mgenericbind   ;
	    m->unbind  = mgenericunbind ;
	    m->remove  = mgenericremove;
	}
    } else {
	/* this is a map with variable length external id's */
	m->resolve = mgenericerror;
	m->bind =  (XMapBindFunc)mgenericerror;
	m->unbind = mgenericunbind;
	m->remove = mgenericremove;
    }
    return m;
} /* mapCreate */


static xkern_return_t
mgenericerror(table, extptr, intptr)
     Map table;
     VOID *extptr;
     VOID **intptr;
{
    xTrace0(idmap, TR_ERRORS,
	    "variable length map accessed via mapBind or mapResolve");
    xTrace0(idmap, TR_ERRORS,
	    "use mapVarBind or mapVarResolve instead");
    Kabort("wrong map function");
    return XK_FAILURE;
} /* mgenericerror */


xkern_return_t
mapVarUnbind( table, ext, len )
    Map table;
    register VOID *ext;
    int len;
{
    Bind     elem_posn, *prev_elem;
    register char *o_ext;
    int		ind;

    ind = generichash((char *)ext, table->tableSize, len);
    prev_elem = &table->table[ind];
    elem_posn = *prev_elem;
    table->cache = 0;
    while (elem_posn != 0) {
	o_ext = (char *)elem_posn->externalid;
	if (! bcmp(o_ext, ext, len) ) {
	    *prev_elem = elem_posn->next;
	    FREEVARMAPELEM(table, elem_posn);
	    return XK_SUCCESS;
	} 
	prev_elem = &(elem_posn->next);
	elem_posn = elem_posn->next;
    }
    return XK_FAILURE;
} /* mapVarUnbind */


Bind
mapVarBind(table, ext, len, intern)
     Map table;
     register VOID *ext;
     int len;
     VOID *intern;
{
    Bind elem_posn, new_elem, prev_elem, *table_posn;
    register char *o_ext;
    int ind;

    ind = generichash((char *)ext, table->tableSize, len);
    table_posn = &table->table[ind];

    elem_posn = *table_posn;
    prev_elem = elem_posn;
    while (elem_posn != 0) {
	o_ext = elem_posn->externalid;
	if (!bcmp(o_ext, (char *)ext, len)) {
	    if (elem_posn->internalid == intern) {
		return(elem_posn);
	    } else {
		return(ERR_BIND);
	    }
	} else {
	    prev_elem = elem_posn;
	    elem_posn = elem_posn->next;
	}
    }
    /*
     * Elements must be inserted at the end of the list rather than at the
     * beginning for the semantics mapForEach to work properly.
     */
    GETVARMAPELEM(table, new_elem, len);
    bcopy((char *)ext, (char *)new_elem->externalid, len);
    new_elem->internalid = intern;
    new_elem->next = 0;
    new_elem->elmlen = len;
    table->cache = new_elem;
    if (prev_elem == 0) {
	*table_posn = new_elem;
    } else {
	prev_elem->next = new_elem;
    }
    return new_elem;
} /* mapVarBind */


xkern_return_t
mapVarResolve(table, ext, len, resPtr)
     Map table;
     register VOID *ext;
     int len;
     register VOID **resPtr;
{
    register Bind elem_posn;
    register char *o_ext;

    elem_posn = table->cache;
    if (elem_posn && elem_posn->elmlen == len) {
	o_ext = elem_posn->externalid;
	if (!bcmp(o_ext, (char *)ext, len)) {
	    if (resPtr) {
		*resPtr = elem_posn->internalid;
	    }
	    return XK_SUCCESS;
	}
    }
    elem_posn = table->table[generichash(ext, table->tableSize, len)];
    while (elem_posn != 0) {
	o_ext = elem_posn->externalid;
	if (elem_posn->elmlen == len && !bcmp(o_ext, (char *)ext, len)) {
	    table->cache = elem_posn;
	    if (resPtr) {
		*resPtr = elem_posn->internalid;
	    }
	    return XK_SUCCESS;
	} else {
	    elem_posn = elem_posn->next;
	}
    }
    return XK_FAILURE;
} /* mapVarResolve */


static void
removeList(e)
     MapElement	*e;
{
    MapElement *next;

    while (e) {
	xTrace1(idmap, TR_FULL_TRACE, "mapClose removing element %lx",
		(unsigned long)e);
	next = e->next;
	xFree((char *)e);
	e = next;
    }
} /* removeList */


void
mapClose(m)
     Map m;
{
    int i;

    xTrace1(idmap, TR_MAJOR_EVENTS, "mapClose of map %lx", (unsigned long) m);
    /* 
     * Free any elements still bound into the map
     */
    for ( i=0; i < m->tableSize; i++ ) {
	removeList(m->table[i]);
    }
    /* 
     * Remove freelist
     */
    removeList(m->freelist);
    xFree((char *)m->table);
    xFree((char *)m);
} /* mapClose */


void
mapForEach(m, f, arg)
     Map m;
     MapForEachFun f;
     VOID *arg;
{
    char	key[MAX_MAP_KEY_SIZE];
    MapElement	*elem, *prevElem;
    int 	i;
    int		userResult = 0;

    xTrace0(idmap, TR_FULL_TRACE, "mapForEach called");
    for (i = 0; i < m->tableSize; i++) {
	prevElem = 0;
	do {
	    /* 
	     * Set elem to first item in the bucket
	     */
	    elem = m->table[i];
	    if (prevElem) {
		/*
		 * Set elem to the next element after the old elem in the
		 * same hash bucket.  If the previous element has been
		 * removed we will skip the rest of this bucket.  If
		 * the previous element has been removed and
		 * reinserted (with a possibly different key/value),
		 * we will skip it and everything before it in the
		 * list. 
		 */
		for ( ; elem != 0 ; elem = elem->next ) {
		    if (prevElem == elem) {
			elem = elem->next;
			break;
		    }
		}
		if (userResult & MFE_REMOVE) {
		    m->remove(m, prevElem);
		}
		prevElem = 0;
	    }
	    if (elem != 0) {
		bcopy((char *)elem->externalid, key, m->keySize);
		userResult = f(key, elem->internalid, arg);
		if ( !(userResult & MFE_CONTINUE) ) {
		    return;
		}
		prevElem = elem;
	    }
	} while ( elem != 0 );
    }
} /* mapForEach */

#include "idmap_special.h"

void
map_init()
{
#define KEYSIZE 2
#define BINDNAME m2bind
#define REMOVENAME m2remove
#define RESOLVENAME m2resolve
#define UNBINDNAME m2unbind
#include "idmap_init.c"
#define KEYSIZE 4
#define BINDNAME m4bind
#define REMOVENAME m4remove
#define RESOLVENAME m4resolve
#define UNBINDNAME m4unbind
#include "idmap_init.c"
#define KEYSIZE 6
#define BINDNAME m6bind
#define REMOVENAME m6remove
#define RESOLVENAME m6resolve
#define UNBINDNAME m6unbind
#include "idmap_init.c"
#define KEYSIZE 8
#define BINDNAME m8bind
#define REMOVENAME m8remove
#define RESOLVENAME m8resolve
#define UNBINDNAME m8unbind
#include "idmap_init.c"
#define KEYSIZE 10
#define BINDNAME m10bind
#define REMOVENAME m10remove
#define RESOLVENAME m10resolve
#define UNBINDNAME m10unbind
#include "idmap_init.c"
#define KEYSIZE 12
#define BINDNAME m12bind
#define REMOVENAME m12remove
#define RESOLVENAME m12resolve
#define UNBINDNAME m12unbind
#include "idmap_init.c"
#define KEYSIZE 16
#define BINDNAME m16bind
#define REMOVENAME m16remove
#define RESOLVENAME m16resolve
#define UNBINDNAME m16unbind
#include "idmap_init.c"
} /* map_init */

			/*** end of idmap.c ***/
