/*
 * @COPYRIGHT@
 * 
 * x-kernel v3.2
 * 
 * Copyright (c) 1993,1991,1990  Arizona Board of Regents
 * 
 * @COPYRIGHT@
 *
 * $RCSfile: _doprint.c,v $
 *
 * HISTORY
 * $Log: _doprint.c,v $
 * Revision 1.5  1994/06/11  00:03:08  menze
 * Added copyright notice
 *
 * Revision 1.4  1994/05/14  23:47:29  davidm
 * Various fixes.  Major change: don't divide by non-constants to avoid
 * library call!
 *
 * Revision 1.3  1994/01/21  19:43:06  davidm
 * (_doarg): Added support for variable width/precision specifier '*' as it
 * is used in the message tool.
 *
 * Revision 1.2  1994/01/18  05:52:39  davidm
 * (doarg): %p fixed to be equivalent to %lx (not %lX).
 *
 * Revision 1.1  1994/01/18  05:23:46  davidm
 * Initial revision
 *
 */
/*
 * Subset of ANSI C printf() formatting.  The major limitation is that
 * floating point number support is non-existant.
 */
#include <stdarg.h>
#include <sys/types.h>

#define FMT_LONG	(1<<0)	/* long value? */
#define FMT_SIGNED	(1<<1)	/* signed value? */
#define FMT_FORCE_SIGN	(1<<2)	/* force printing of "sign"? */
#define FMT_PLUS_SPACE	(1<<3)	/* use space instead of plus sign? */
#define FMT_LEFT_ALIGN	(1<<4)	/* left aligned? */
#define FMT_ZERO_PAD	(1<<5)	/* zero padding instead of space padding? */
#define FMT_ALTERNATE	(1<<6)	/* prefix 0 to %o, 0x to %x? */
#define FMT_CAPITAL	(1<<7)	/* use capital letters for hex */


#define	MAXBITS	(8*sizeof(void*))


static char*
format_int(char *buf, int buflen,
	   unsigned long int v, int base, int flags, int width, int prec,
	   int *lenp)
{
    int digit;
    char *cp = &buf[buflen - 1];
    int positive = 1;
    int len = 0;

    *cp-- = '\0';
    if (flags & FMT_SIGNED) {
	if (flags & FMT_LONG) {
	    if ((long int) v < 0) {
		v = -(long int) v;
		positive = 0;
	    } /* if */
	} else {
	    if ((int) v < 0) {
		v = -(long)(int) v;
		positive = 0;
	    } /* if */
	} /* if */
    } /* if */

    /* format number: */
    do {
	/* avoid divison---it's a library call! */
	switch (base) {
	  case  8: digit = v & 0x7;  v >>= 3; break;
	  case 10: digit = v % 10;   v /= 10; break;
	  case 16: digit = v & 0xf;  v >>= 4; break;
	  default: digit = v % base; v /= base; break;
	} /* switch */

	if (digit <= 9) {
	    *cp-- = '0' + digit;
	} else {
	    if (flags & FMT_CAPITAL) {
		*cp-- = 'A' + digit - 10;
	    } else {
		*cp-- = 'a' + digit - 10;
	    } /* if */
	} /* if */
	++len;
    } while (v);

    if ((flags & FMT_ZERO_PAD) && (prec < 0) && !(flags & FMT_LEFT_ALIGN)) {
	prec = width;
	if ((flags & FMT_SIGNED) && ((flags & FMT_FORCE_SIGN) || !positive)) {
	    --prec;	/* account for sign */
	} /* if */
    } /* if */

    while (len < prec && len < buflen - 4) {
	*cp-- = '0'; ++len;
    } /* while */

    if (flags & FMT_SIGNED) {
	if (positive) {
	    if (flags & FMT_FORCE_SIGN) {
		if (flags & FMT_PLUS_SPACE) {
		    *cp-- = ' '; ++len;
		} else {
		    *cp-- = '+'; ++len;
		} /* if */
	    } /* if */
	} else {
	    *cp-- = '-'; ++len;
	} /* if */
    } /* if */
    *lenp = len;
    return ++cp;
} /* format_int */


static void
align_str(char *str, int str_len, int width, int flags,
	  void (*outc)(int, void *), void *arg, int *countp)
{
    int i;

    if (!(flags & FMT_LEFT_ALIGN)) {
	for (i = 0; i < width - str_len; ++i) {
	    (*outc)(' ', arg); ++(*countp);
	} /* for */
    } /* if */

    for (i = 0; i < str_len; ++i) {
	(*outc)(*str++, arg); ++(*countp);
    } /* for */

    if (flags & FMT_LEFT_ALIGN) {
	for (i = 0; i < width - str_len; ++i) {
	    (*outc)(' ', arg); ++(*countp);
	} /* for */
    } /* if */
} /* align_str */


static const char*
doarg(const char *fmt, va_list *ap, void (*outc)(int, void *), void *arg,
      int *countp)
{
    unsigned long int v;
    unsigned char ch, *str;
    int flags = 0;
#   define S_PAD	0	/* scanning padding */
#   define S_WIDTH	1	/* scanning width */
#   define S_PREC	2	/* scanning precision */
    int state = S_PAD;
    int len, iv;
    int width = 0;		/* field width */
    int prec = -1;		/* precision */
    char buf[MAXBITS+3];

    while ((ch = *fmt++) != '\0') {
	switch (ch) {
	  case '-':
	    flags |= FMT_LEFT_ALIGN;
	    break;

	  case '+':
	    flags |= FMT_FORCE_SIGN;
	    flags &= ~FMT_PLUS_SPACE;
	    break;

	  case ' ':
	    flags |= FMT_FORCE_SIGN | FMT_PLUS_SPACE;
	    break;

	  case '#':
	    flags |= FMT_ALTERNATE;
	    break;

	  case '.':
	    state = S_PREC;
	    break;

	  case '0': case '1': case '2': case '3': case '4':
	  case '5': case '6': case '7': case '8': case '9':
	    switch (state) {
	      case S_PAD:
		state = S_WIDTH;
		if (ch == '0') {
		    flags |= FMT_ZERO_PAD;
		    break;
		} /* if */
		/*
		 * FALL THROUGH
		 */
	      case S_WIDTH:
		width = 0;
		do {
		    width = 10 * width + ch - '0';
		} while ((ch = *fmt++) >= '0' && ch <= '9');
		--fmt;
		break;

	      case S_PREC:
		prec = 0;
		do {
		    prec = 10 * prec + ch - '0';
		} while ((ch = *fmt++) >= '0' && ch <= '9');
		--fmt;
		break;
	    } /* switch */
	    break;

	  case '*':
	    iv = va_arg(*ap, int);
	    if (iv < 0) {
		flags |= FMT_LEFT_ALIGN;
		iv = -iv;
	    } /* if */
	    if (state == S_PREC) {
		prec  = iv;
	    } else {
		width = iv;
	    } /* if */
	    break;

	  case '%':
	    buf[0] = '%';
	    align_str(buf, 1, width, flags, outc, arg, countp);
	    return fmt;

	  case 'c':
	    buf[0] = va_arg(*ap, unsigned char);
	    align_str(buf, 1, width, flags, outc, arg, countp);
	    return fmt;

	  case 's':
	    str = va_arg(*ap, char*);
	    len = strlen(str);
	    if (prec >= 0 && len > prec) {
		len = prec;	/* print at most prec characters */
	    } /* if */
	    align_str(str, len, width, flags, outc, arg, countp);
	    return fmt;

	  case 'l':
	    flags |= FMT_LONG;
	    break;

	  case 'b':
	    if (flags & FMT_LONG) {
		v = va_arg(*ap, unsigned long int);
	    } else {
		v = va_arg(*ap, unsigned int);
	    } /* if */
	    str = format_int(buf, sizeof(buf), v, 2, flags, width, prec, &len);
	    align_str(str, len, width, flags, outc, arg, countp);
	    return fmt;

	  case 'o':
	    if (flags & FMT_LONG) {
		v = va_arg(*ap, unsigned long int);
	    } else {
		v = va_arg(*ap, unsigned int);
	    } /* if */
	    if ((flags & FMT_ALTERNATE) && v) {
		--width;	/* account for leading 0 */
	    } /* if */
	    str = format_int(buf, sizeof(buf), v, 8, flags, width, prec, &len);
	    if ((flags & FMT_ALTERNATE) && v) {
		*--str = '0'; ++len;
		++width;	/* restore proper width */
	    } /* if */
	    align_str(str, len, width, flags, outc, arg, countp);
	    return fmt;

	  case 'd':
	  case 'i':
	    flags |= FMT_SIGNED;
	    /*
	     * FALL THROUGH
	     */
	  case 'u':
	    if (flags & FMT_LONG) {
		v = va_arg(*ap, unsigned long int);
	    } else {
		v = (long int) va_arg(*ap, unsigned int);
	    } /* if */
	    str = format_int(buf, sizeof(buf), v, 10, flags, width, prec,
			     &len);
	    align_str(str, len, width, flags, outc, arg, countp);
	    return fmt;

	  case 'p':
	    flags |= FMT_LONG;
	  case 'X':
	    if (ch != 'p') {
		flags |= FMT_CAPITAL;
	    } /* if */
	  case 'x':
	    if (flags & FMT_LONG) {
		v = va_arg(*ap, unsigned long int);
	    } else {
		v = va_arg(*ap, unsigned int);
	    } /* if */
	    if ((flags & FMT_ALTERNATE) && v) {
		width -= 2;		/* account for leading 0x */
	    } /* if */	    
	    str = format_int(buf, sizeof(buf), v, 16, flags, width, prec,
			     &len);
	    if ((flags & FMT_ALTERNATE) && v) {
		*--str = (flags & FMT_CAPITAL) ? 'X' : 'x';
		*--str = '0';
		len += 2;
		width += 2;		/* restore proper width */
	    } /* if */
	    align_str(str, len, width, flags, outc, arg, countp);
	    return fmt;

	  default:
	    (*outc)(ch, arg); (*countp)++;
	    return fmt;
	} /* switch */
    } /* while */
    return fmt;
} /* doarg */


int
_doprint(const char *format, va_list *ap, void (*outc)(int, void*), void *arg)
{
    char ch;
    int count = 0;

    while ((ch = *format++) != 0) {
	if (ch != '%') {
	    (*outc)(ch, arg); count++;
	} else {
	    format = doarg(format, ap, outc, arg, &count);
	} /* if */
    } /* while */
    return count;
} /* _doprint */

			/*** end of _doprint.c ***/
